package com.jiuqi.mde;

import java.sql.Statement;
import java.sql.Connection;
import java.util.ArrayList;
import java.sql.PreparedStatement;
import com.jiuqi.util.IntValue;
import java.sql.ResultSet;
import java.sql.SQLException;
import com.jiuqi.util.SequencialList;
import com.jiuqi.util.JqLib;
import java.util.List;
import com.jiuqi.util.DBConnect;
import com.jiuqi.util.DuplicationException;

public class MDETable extends MDERecordSet
{
    public static boolean KEEP_FIELDS_ORDER;
    public boolean keepFieldsOrder;
    protected StatementCache stmtCache;
    protected MDEDimValueSetCollection loadedDimValues;
    private static final Integer DEFAULT_INT_VALUE;
    private static final String DEFAULT_STR_VALUE = "_";
    
    static {
        MDETable.KEEP_FIELDS_ORDER = false;
        DEFAULT_INT_VALUE = new Integer(999);
    }
    
    public MDETable(final MDEDataInfo dataInfo) {
        super(dataInfo);
        this.stmtCache = new StatementCache();
        this.loadedDimValues = new MDEDimValueSetCollection();
        this.keepFieldsOrder = MDETable.KEEP_FIELDS_ORDER;
    }
    
    public MDERecord createRecord() {
        return new MDERecord(this.dataInfo);
    }
    
    public final MDERecord newRecord() throws MDEException {
        try {
            return this.newRecord(null);
        }
        catch (final DuplicationException ex) {
            ex.printStackTrace();
            return null;
        }
    }
    
    public final MDERecord newRecord(final MDEDimValueSet dimValues) throws DuplicationException, MDEException {
        return this.newRecord(dimValues, true);
    }
    
    public final MDERecord newRecord(final MDEDimValueSet dimValues, final boolean checkNeedLoad) throws DuplicationException, MDEException {
        MDERecord result = null;
        if (this.getIsUseLoadDataSync()) {
            result = this.dvFind(dimValues);
            if (result != null) {
                return result;
            }
            synchronized (this.mutexLoadData) {
                result = this.dvFind(dimValues);
                if (result == null) {
                    result = this.newRecord0(dimValues, checkNeedLoad);
                }
                monitorexit(this.mutexLoadData);
                return result;
            }
        }
        result = this.newRecord0(dimValues, checkNeedLoad);
        return result;
    }
    
    private MDERecord newRecord0(final MDEDimValueSet dimValues, final boolean checkNeedLoad) throws MDEException, DuplicationException {
        final MDERecord result = this.createRecord();
        result.beforeNewRecord();
        if (dimValues != null) {
            result.setDimValues(dimValues, null);
        }
        result.afterNewRecord();
        if (dimValues != null && checkNeedLoad && !this.dataInfo.isTemporary() && this.needLoadData(result.getDimValues())) {
            System.out.println(String.valueOf(this.dataInfo.getName()) + "应该首先载入数据，否则无法定位记录序号");
        }
        try {
            this.add(result);
        }
        catch (final DuplicationException ex) {
            if (!this.hasOrder) {
                throw ex;
            }
            result.getDimValues().setValue("顺序号", null);
            this.add(result);
        }
        result.calculateDefaultValues();
        return result;
    }
    
    public final List getRecords(final DBConnect dbConnect, final MDEDimValueSet dimValues) throws MDEException {
        this.loadData(dbConnect, dimValues);
        return this.dvFindMatches(dimValues);
    }
    
    public final MDERecord getRecord(final DBConnect dbConnect, MDEDimValueSet dimValues, final boolean createIfNotExist) throws MDEException {
        if (!dimValues.dimensions().equals(this.dataInfo.dimensions())) {
            final MDEDimValueSet newDimValues = new MDEDimValueSet(this.dataInfo.dimensions());
            newDimValues.assignValues(dimValues);
            dimValues = newDimValues;
        }
        MDERecord result = this.dvFind(dimValues);
        if (result == null) {
            final List records = this.getRecords(dbConnect, dimValues);
            if (records != null && records.size() > 1) {
                throw new MDEException(String.valueOf(dimValues.toString()) + "维度编码不全，不能唯一定位" + this.dataInfo.getName() + "记录");
            }
            if (records != null && records.size() > 0) {
                result = records.get(0);
            }
            if (result == null) {
                if (!createIfNotExist) {
                    if (!this.dataInfo.isTemporary()) {
                        return result;
                    }
                }
                try {
                    result = this.newRecord(dimValues, false);
                }
                catch (final DuplicationException ex) {
                    ex.printStackTrace();
                }
            }
        }
        return result;
    }
    
    public final boolean needLoadData(final MDEDimValueSet dimValues) {
        return !this.loadedDimValues.isIncluded(dimValues);
    }
    
    public void dataLoaded(final MDEDimValueSet dimValues) {
        this.loadedDimValues.include(dimValues);
    }
    
    public void loadData(final DBConnect dbConnect, MDEDimValueSet dimValues) throws MDEException {
        if ((dimValues.dimensions().collectCompare(this.dataInfo.dimensions()) & 0x1) != 0x0) {
            final MDEDimValueSet myDimValues = new MDEDimValueSet(this.dataInfo.dimensions());
            myDimValues.assignValues(dimValues);
            dimValues = myDimValues;
            dimValues.setValue("顺序号", null);
        }
        else if (dimValues.dimensions().hasOrder() && dimValues.getValue("顺序号") != null) {
            dimValues = (MDEDimValueSet)dimValues.clone();
            dimValues.setValue("顺序号", null);
        }
        if (!this.needLoadData(dimValues)) {
            return;
        }
        if (this.getIsUseLoadDataSync()) {
            synchronized (this.mutexLoadData) {
                if (!this.needLoadData(dimValues)) {
                    monitorexit(this.mutexLoadData);
                    return;
                }
                this.loadData0(dbConnect, dimValues);
                monitorexit(this.mutexLoadData);
                return;
            }
        }
        this.loadData0(dbConnect, dimValues);
    }
    
    private void dealDuplication(final MDERecord record) throws MDEException {
        final MDERecord oldRecord = this.dvFind(record.getDimValues());
        final MDEField field = this.dataInfo.getIndentyField();
        if (field != null && JqLib.compareString((String)record.getFieldValue(field), (String)oldRecord.getFieldValue(field)) != 0) {
            if (!this.dataInfo.dimensions().hasOrder()) {
                System.out.println(String.valueOf(this.dataInfo.getName()) + "数据库记录重码：" + record.getDimValues().toString() + "，记录被忽略");
            }
            else {
                System.out.println(String.valueOf(this.dataInfo.getName()) + "数据库记录重码：" + record.getDimValues().toString() + "，调整顺序号");
                record.getDimValues().setValue("顺序号", null);
                try {
                    this.add(record);
                }
                catch (final DuplicationException ex) {}
                record.setDimValues(record.getDimValues(), null);
            }
        }
        else {
            if ((oldRecord.recordFlag & 0x2) != 0x0) {
                throw new MDEException(String.valueOf(this.dataInfo.getName()) + "记录已经修改，不能再次从数据库载入");
            }
            try {
                this.remove(oldRecord);
                this.add(record);
            }
            catch (final DuplicationException ex2) {}
        }
    }
    
    protected void add0(final MDERecord record) throws MDEException {
        if (this.dimValuesIndex == null) {
            this.buildDimValuesIndex();
        }
        try {
            this.add(record);
        }
        catch (final DuplicationException ex) {
            this.dealDuplication(record);
        }
    }
    
    private void loadData0(final DBConnect dbConnect, final MDEDimValueSet dimValues) throws MDEException {
        if (this.dimValuesIndex == null) {
            this.buildDimValuesIndex();
        }
        final SequencialList result = this.loadFromTable(dbConnect, dimValues);
        for (int index = 0, count = result.size(); index < count; ++index) {
            final MDERecord record = (MDERecord)result.get(index);
            try {
                this.add(record);
            }
            catch (final DuplicationException ex) {
                this.dealDuplication(record);
            }
        }
        this.dataLoaded(dimValues);
    }
    
    public final void statModifies(final List modifyRecrods, final MDEDimValueSetCollection modifyDimValues) {
        for (int i = 0, count = this.records.size(); i < count; ++i) {
            final MDERecord record = this.records.get(i);
            if (record.isInserted() || record.isDeleted() || record.isModified()) {
                if (modifyRecrods != null) {
                    modifyRecrods.add(record);
                }
                if (modifyDimValues != null) {
                    modifyDimValues.add(record.getDimValues());
                }
            }
        }
    }
    
    public final void saveData(final DBConnect dbConnect) throws MDEException {
        try {
            dbConnect.getConnection();
            try {
                this.stmtCache.clear();
                try {
                    for (int i = 0, count = this.records.size(); i < count; ++i) {
                        final MDERecord record = this.records.get(i);
                        if (record.isInserted()) {
                            record.refreshDimValues(this.dimValuesIndex);
                        }
                        if (!record.isDeleting() && (record.isModified() || record.isInserted()) && !record.hasData()) {
                            record.delete();
                        }
                    }
                    for (int i = this.records.size() - 1; i >= 0; --i) {
                        final MDERecord record2 = this.records.get(i);
                        if (record2.isDeleting()) {
                            this.deleteRecord(dbConnect, record2);
                            this.remove(i);
                        }
                    }
                    for (int i = 0, count = this.records.size(); i < count; ++i) {
                        final MDERecord record = this.records.get(i);
                        if (record.isInserted()) {
                            this.insertRecord(dbConnect, record);
                        }
                        else if (!record.isDeleting()) {
                            record.checkModification();
                            if (record.isModified()) {
                                this.updateRecord(dbConnect, record);
                            }
                        }
                    }
                }
                finally {
                    this.stmtCache.close();
                }
                this.stmtCache.close();
            }
            finally {
                dbConnect.releaseConnection();
            }
            dbConnect.releaseConnection();
        }
        catch (final SQLException ex) {
            throw new MDEException(ex);
        }
    }
    
    @Override
    public void resetRecords() {
        this.loadedDimValues.clear();
        super.resetRecords();
    }
    
    protected Object readDimension(final ResultSet resultSet, final MDEDimension dimension, final IntValue index) throws SQLException {
        if ("顺序号" == dimension.name) {
            String order = null;
            if (index == null) {
                order = resultSet.getString(dimension.fieldName);
            }
            else {
                order = resultSet.getString(index.value++);
            }
            try {
                return new Integer(JqLib.parseHexValue(order));
            }
            catch (final NumberFormatException ex) {
                System.out.println(String.valueOf(this.dataInfo.getName()) + "顺序号无效");
                ex.printStackTrace();
                return new Integer(1);
            }
        }
        if (index == null) {
            if (dimension.isString) {
                return resultSet.getString(dimension.fieldName);
            }
            return new Integer(resultSet.getInt(dimension.fieldName));
        }
        else {
            if (dimension.isString) {
                return resultSet.getString(index.value++);
            }
            return new Integer(resultSet.getInt(index.value++));
        }
    }
    
    protected void writeDimension(final PreparedStatement statement, final MDEDimension dimension, final IntValue index, final Object value) throws SQLException {
        if ("顺序号" == dimension.name) {
            final int order = (int)((value != null) ? value : 1);
            statement.setString(index.value++, JqLib.makeHexString(order));
        }
        else if (dimension.isString) {
            statement.setString(index.value++, (String)value);
        }
        else if (value instanceof Integer) {
            if (value == null) {
                statement.setNull(index.value++, 4);
            }
            else {
                statement.setInt(index.value++, (int)value);
            }
        }
        else if (value == null) {
            statement.setNull(index.value++, 2);
        }
        else {
            statement.setDouble(index.value++, (double)value);
        }
    }
    
    protected Object readField(final ResultSet resultSet, final MDEField field, final IntValue index) throws SQLException {
        if (index == null) {
            return resultSet.getObject(field.name);
        }
        return resultSet.getObject(index.value++);
    }
    
    protected void writeField(final PreparedStatement statement, final MDEField field, final IntValue index, final Object value) throws SQLException {
        if (value == null) {
            statement.setNull(index.value++, field.dataType);
        }
        else {
            statement.setObject(index.value++, value);
        }
    }
    
    protected final void readRecord(final ResultSet resultSet, final MDERecord record, final boolean readDimension) throws SQLException {
        IntValue index = null;
        if (MDETable.KEEP_FIELDS_ORDER) {
            index = new IntValue(1);
        }
        if (readDimension) {
            final MDEDimValueSet dimValues = new MDEDimValueSet(this.dataInfo.dimensions());
            for (int i = 0, count = this.dataInfo.dimensions().size(); i < count; ++i) {
                final MDEDimension dimension = this.dataInfo.dimensions().get(i);
                dimValues.setValue(i, this.readDimension(resultSet, dimension, index));
            }
            record.setDimValues(dimValues, null);
        }
        for (int j = 0, count2 = this.fields.size(); j < count2; ++j) {
            final MDEField field = this.fields.get(j);
            record.setFieldValue(j, this.readField(resultSet, field, index));
        }
    }
    
    protected final void writeFieldsData(final PreparedStatement statement, final MDERecord record, final IntValue index) throws SQLException {
        for (int i = 0, count = this.fields.size(); i < count; ++i) {
            final MDEField field = this.fields.get(i);
            this.writeField(statement, field, index, record.getFieldValue(i));
        }
    }
    
    protected final void writeDimensionsData(final PreparedStatement statement, final MDERecord record, final IntValue index, final boolean writeFieldDimension) throws SQLException {
        final MDEDimValueSet dimValues = record.getDimValues();
        for (int i = 0, count = this.dataInfo.dimensions().size(); i < count; ++i) {
            final MDEDimension dimension = this.dataInfo.dimensions().get(i);
            if (writeFieldDimension || this.fields.findByFieldName(dimension.fieldName) == null) {
                this.writeDimension(statement, dimension, index, dimValues.getValue(i));
            }
        }
    }
    
    protected final void writeRecord(final PreparedStatement statement, final MDERecord record, final boolean writeDimension) throws SQLException {
        final IntValue index = new IntValue(1);
        if (writeDimension) {
            final MDEDimValueSet dimValues = record.getDimValues();
            for (int i = 0, count = this.dataInfo.dimensions().size(); i < count; ++i) {
                final MDEDimension dimension = this.dataInfo.dimensions().get(i);
                if (this.fields.findByFieldName(dimension.fieldName) == null) {
                    this.writeDimension(statement, dimension, index, dimValues.getValue(i));
                }
            }
        }
        for (int j = 0, count2 = this.fields.size(); j < count2; ++j) {
            final MDEField field = this.fields.get(j);
            this.writeField(statement, field, index, record.getFieldValue(j));
        }
    }
    
    private final void checkEmptyFields(final MDEDimValueSet dimValues, final ArrayList emptyFields) {
        for (int i = 0, count = dimValues.dimensions().size(); i < count; ++i) {
            final Object dimValue = dimValues.getValue(i);
            if (dimValue == null || "".equals(dimValue)) {
                final MDEDimension dim = dimValues.dimensions().get(i);
                if (emptyFields.indexOf(dim) < 0) {
                    emptyFields.add(dim);
                }
            }
        }
    }
    
    private final void convertEmptyDimension(final Connection conn, final MDEDimension dim) throws SQLException {
        final StringBuffer sql = new StringBuffer(80);
        sql.append("UPDATE ");
        sql.append(this.getTableName());
        sql.append(" SET ");
        sql.append(dim.fieldName);
        if (dim.isString) {
            sql.append("='_'");
        }
        else {
            sql.append("=99");
        }
        sql.append(" WHERE ");
        sql.append(dim.fieldName);
        if (dim.isString) {
            sql.append("= '' OR ");
        }
        else {
            sql.append("=0 OR ");
        }
        sql.append(dim.fieldName);
        sql.append(" IS NULL");
        final Statement stmt = conn.createStatement();
        try {
            stmt.executeUpdate(sql.toString());
        }
        finally {
            stmt.close();
        }
        stmt.close();
    }
    
    private final void changeEmptyDimensionValue(final SequencialList records, final MDEDimension dim) {
        for (int index = 0, count = records.size(); index < count; ++index) {
            final MDERecord rec = (MDERecord)records.get(index);
            final MDEDimValueSet dimValues = rec.getDimValues();
            if (dimValues.getValue(dim) == null) {
                if (dim.isString) {
                    dimValues.setValue(dim, "_");
                }
                else {
                    dimValues.setValue(dim, MDETable.DEFAULT_INT_VALUE);
                }
            }
        }
    }
    
    protected String convertSQL(final String sql, final MDEDimValueSet dimValues) {
        return sql;
    }
    
    protected void doLoadRecord(final ResultSet rs, final MDERecord record) throws SQLException, MDEException {
        record.beforeLoadRecord();
        this.readRecord(rs, record, true);
        record.afterLoadRecord();
    }
    
    private String generateSql(final MDEDimValueSet dimValues) {
        return this.generateSql(dimValues, false);
    }
    
    private String generateSql(final MDEDimValueSet dimValues, final boolean isPrepared) {
        StringBuffer sql = null;
        if (!this.keepFieldsOrder) {
            sql = new StringBuffer(128);
            sql.append("SELECT * FROM ");
        }
        else {
            sql = new StringBuffer(this.fields.size() * 8 + 80);
            sql.append("SELECT ");
            for (int i = 0, count = this.dataInfo.dimensions().size(); i < count; ++i) {
                sql.append(this.dataInfo.dimensions().get(i).fieldName);
                sql.append(',');
            }
            for (int i = 0, count = this.fields.size(); i < count; ++i) {
                final MDEField field = this.fields.get(i);
                if (this.dataInfo.dimensions().findByFieldName(field.fieldName) == null) {
                    sql.append(field.fieldName);
                    sql.append(',');
                }
            }
            sql.deleteCharAt(sql.length() - 1);
            sql.append(" FROM ");
        }
        sql.append(this.getTableName());
        sql.append(" WHERE ");
        sql.append(dimValues.getSQLFilter(isPrepared));
        return sql.toString();
    }
    
    protected void beforeReadRecord(final MDEDimValueSet dimValues) {
    }
    
    protected void readRecord(final ResultSet rs, final SequencialList result, final MDERecord record) throws SQLException, MDEException {
        this.doLoadRecord(rs, record);
        result.add((Object)record);
    }
    
    protected void afterReadRecord(final MDEDimValueSet dimValues) {
    }
    
    protected SequencialList loadFromTable(final DBConnect dbConnect, final MDEDimValueSet dimValues) throws MDEException {
        final SequencialList result = new SequencialList();
        boolean isPrepared = false;
        final String sql = this.generateSql(dimValues);
        try {
            final Connection conn = dbConnect.getConnection();
            try {
                final ArrayList emptyFields = new ArrayList();
                PreparedStatement stmt = null;
                try {
                    this.beforeReadRecord(dimValues);
                    String selectSQL = this.convertSQL(sql, dimValues);
                    if (selectSQL.equals(sql)) {
                        isPrepared = true;
                        selectSQL = this.generateSql(dimValues, isPrepared);
                    }
                    stmt = conn.prepareStatement(selectSQL);
                    if (isPrepared) {
                        this.setParameter(stmt, dimValues);
                    }
                    final ResultSet rs = stmt.executeQuery();
                    try {
                        while (rs.next()) {
                            final MDERecord record = this.createRecord();
                            this.readRecord(rs, result, record);
                            this.checkEmptyFields(record.getDimValues(), emptyFields);
                        }
                        this.afterReadRecord(dimValues);
                    }
                    finally {
                        rs.close();
                    }
                    rs.close();
                    if (emptyFields.size() > 0) {
                        for (int i = 0, count = emptyFields.size(); i < count; ++i) {
                            final MDEDimension dim = emptyFields.get(i);
                            this.convertEmptyDimension(conn, dim);
                            this.changeEmptyDimensionValue(result, dim);
                        }
                    }
                }
                finally {
                    stmt.close();
                }
                stmt.close();
            }
            finally {
                dbConnect.releaseConnection();
            }
            dbConnect.releaseConnection();
        }
        catch (final SQLException ex) {
            throw new MDEException(ex);
        }
        return result;
    }
    
    private void setParameter(final PreparedStatement ps, final MDEDimValueSet dimValues) throws SQLException {
        if (dimValues == null) {
            return;
        }
        final MDEDimensionSet dimensions = dimValues.dimensions();
        final Object[] values = dimValues.getDimValues();
        int index = 1;
        if (dimensions != null) {
            for (int i = 0; i < dimensions.size(); ++i) {
                final MDEDimension dimension = dimensions.get(i);
                final Object value = values[i];
                if (value != null) {
                    index = dimension.setParameter(ps, value, index);
                }
            }
        }
    }
    
    protected final boolean insertRecord(final DBConnect dbConnect, final MDERecord record) throws MDEException {
        try {
            int result = 0;
            record.beforeSaveRecord();
            try {
                result = this.doInsertRecord(dbConnect, record);
            }
            catch (final SQLException e) {
                if (!DBConnect.isUniqueViolationException(e)) {
                    throw e;
                }
                result = this.doUpdateRecord(dbConnect, record);
            }
            record.afterSaveRecord();
            if (MDESqlLogger.Sql_Logger != null) {
                MDESqlLogger.Sql_Logger.LogInsertSql(this.makeInsertInfo(record, result));
            }
            return result > 0;
        }
        catch (final SQLException ex) {
            throw new MDEException(ex);
        }
    }
    
    private final int doInsertRecord(final DBConnect dbConnect, final MDERecord record) throws SQLException {
        String sqlString = null;
        PreparedStatement stmt = this.stmtCache.insertStatement;
        if (stmt == null) {
            final StringBuffer sql = new StringBuffer(this.fields.size() * 10 + 100);
            final StringBuffer param = new StringBuffer(20 + 2 * this.fields.size());
            sql.append("INSERT INTO ");
            sql.append(this.getTableName());
            sql.append(" (");
            for (int i = 0, count = this.dataInfo.dimensions().size(); i < count; ++i) {
                final MDEDimension dimension = this.dataInfo.dimensions().get(i);
                if (this.fields.findByFieldName(dimension.fieldName) == null) {
                    dimension.appendInsertSql(sql, param);
                }
            }
            for (int i = 0, count = this.fields.size(); i < count; ++i) {
                sql.append(this.fields.get(i).fieldName);
                sql.append(',');
                param.append("?,");
            }
            sql.deleteCharAt(sql.length() - 1);
            param.deleteCharAt(param.length() - 1);
            sql.append(")");
            sql.append(" VALUES (");
            sql.append(param);
            sql.append(")");
            sqlString = this.convertSQL(sql.toString(), record.getDimValues());
        }
        final Connection conn = dbConnect.getConnection();
        try {
            if (stmt == null) {
                stmt = conn.prepareStatement(sqlString);
                this.stmtCache.insertStatement = stmt;
            }
            final IntValue index = new IntValue(1);
            this.writeDimensionsData(stmt, record, index, false);
            this.writeFieldsData(stmt, record, index);
            return stmt.executeUpdate();
        }
        finally {
            dbConnect.releaseConnection();
        }
    }
    
    protected final boolean updateRecord(final DBConnect dbConnect, final MDERecord record) throws MDEException {
        try {
            int result = 0;
            record.beforeSaveRecord();
            result = this.doUpdateRecord(dbConnect, record);
            if (result == 0) {
                result = this.doInsertRecord(dbConnect, record);
            }
            record.afterSaveRecord();
            if (MDESqlLogger.Sql_Logger != null) {
                MDESqlLogger.Sql_Logger.LogInsertSql(this.makeUpdateInfo(record, result));
            }
            return result > 0;
        }
        catch (final SQLException ex) {
            throw new MDEException(ex);
        }
    }
    
    private final int doUpdateRecord(final DBConnect dbConnect, final MDERecord record) throws SQLException {
        final MDEField identyField = this.dataInfo.getIndentyField();
        final Object identyValue = (identyField == null) ? null : record.getFieldValue(identyField);
        String sqlString = null;
        PreparedStatement stmt = (identyValue == null) ? this.stmtCache.updateStatement : this.stmtCache.updateByIdentyStatement;
        if (stmt == null) {
            final StringBuffer sql = new StringBuffer(this.fields.size() * 10 + 100);
            sql.append("UPDATE ");
            sql.append(this.getTableName());
            sql.append(" SET ");
            for (int i = 0, count = this.fields.size(); i < count; ++i) {
                sql.append(this.fields.get(i).fieldName);
                sql.append('=');
                sql.append("?,");
            }
            sql.deleteCharAt(sql.length() - 1);
            sql.append(" WHERE ");
            if (identyValue != null) {
                sql.append(identyField.fieldName);
                sql.append("=?");
            }
            else {
                for (int i = 0; i < this.dataInfo.dimensions().size(); ++i) {
                    final MDEDimension dimension = this.dataInfo.dimensions().get(i);
                    dimension.appendWhereClause(sql);
                }
                sql.setLength(sql.length() - 4);
            }
            sqlString = this.convertSQL(sql.toString(), record.getDimValues());
        }
        final Connection conn = dbConnect.getConnection();
        try {
            if (stmt == null) {
                stmt = conn.prepareStatement(sqlString);
                if (identyValue != null) {
                    this.stmtCache.updateByIdentyStatement = stmt;
                }
                else {
                    this.stmtCache.updateStatement = stmt;
                }
            }
            final IntValue index = new IntValue(1);
            this.writeFieldsData(stmt, record, index);
            if (identyValue == null) {
                this.writeDimensionsData(stmt, record, index, true);
            }
            else {
                this.writeField(stmt, identyField, index, identyValue);
            }
            return stmt.executeUpdate();
        }
        finally {
            dbConnect.releaseConnection();
        }
    }
    
    protected final boolean deleteRecord(final DBConnect dbConnect, final MDERecord record) throws MDEException {
        try {
            record.beforeDeleteRecord();
            final int result = this.doDeleteRecord(dbConnect, record);
            record.afterDeleteRecord();
            if (MDESqlLogger.Sql_Logger != null) {
                MDESqlLogger.Sql_Logger.LogInsertSql(this.makeDeleteInfo(record.getDimValues(), result));
            }
            return result > 0;
        }
        catch (final SQLException ex) {
            throw new MDEException(ex);
        }
    }
    
    private final int doDeleteRecord(final DBConnect dbConnect, final MDERecord record) throws SQLException {
        final MDEField identyField = this.dataInfo.getIndentyField();
        final Object identyValue = (identyField == null) ? null : record.getFieldValue(identyField);
        String sqlString = null;
        PreparedStatement stmt = (identyValue == null) ? this.stmtCache.deleteStatement : this.stmtCache.deleteByIdentyStatement;
        if (stmt == null) {
            final StringBuffer sql = new StringBuffer(this.fields.size() * 10 + 100);
            sql.append("DELETE FROM ");
            sql.append(this.getTableName());
            sql.append(" WHERE ");
            if (identyValue != null) {
                sql.append(identyField.fieldName);
                sql.append("=?");
            }
            else {
                for (int i = 0; i < this.dataInfo.dimensions().size(); ++i) {
                    final MDEDimension dimension = this.dataInfo.dimensions().get(i);
                    dimension.appendWhereClause(sql);
                }
                sql.setLength(sql.length() - 4);
            }
            sqlString = this.convertSQL(sql.toString(), record.getDimValues());
        }
        final Connection conn = dbConnect.getConnection();
        try {
            if (stmt == null) {
                stmt = conn.prepareStatement(sqlString);
                if (identyValue != null) {
                    this.stmtCache.deleteByIdentyStatement = stmt;
                }
                else {
                    this.stmtCache.deleteStatement = stmt;
                }
            }
            final IntValue index = new IntValue(1);
            if (identyValue == null) {
                this.writeDimensionsData(stmt, record, index, true);
            }
            else {
                this.writeField(stmt, identyField, index, identyValue);
            }
            return stmt.executeUpdate();
        }
        finally {
            dbConnect.releaseConnection();
        }
    }
    
    public int deleteRecords(final DBConnect dbConnect, MDEDimValueSet dimValues) throws SQLException {
        if (!dimValues.dimensions().equals(this.dataInfo.dimensions())) {
            final MDEDimValueSet newDimValues = new MDEDimValueSet(this.dataInfo.dimensions());
            newDimValues.assignValues(dimValues);
            dimValues = newDimValues;
        }
        final String filter = dimValues.getSQLFilter();
        if (filter == null || filter.length() == 0) {
            return 0;
        }
        final Connection conn = dbConnect.getConnection();
        try {
            final Statement stmt = conn.createStatement();
            try {
                final StringBuffer sql = new StringBuffer(128);
                sql.append("DELETE FROM ");
                sql.append(this.getTableName());
                sql.append(" WHERE ");
                sql.append(filter);
                final int result = stmt.executeUpdate(this.convertSQL(sql.toString(), dimValues));
                if (MDESqlLogger.Sql_Logger != null) {
                    MDESqlLogger.Sql_Logger.LogInsertSql(this.makeDeleteInfo(dimValues, result));
                }
                return result;
            }
            finally {
                stmt.close();
            }
        }
        finally {
            dbConnect.releaseConnection();
        }
    }
    
    protected String getTableName() {
        return this.dataInfo.getName();
    }
    
    private void writeFieldValues(final MDERecord record, final StringBuffer sb) {
        for (int i = 0, count = this.fields.size(); i < count; ++i) {
            final MDEField field = this.fields.get(i);
            sb.append('\t');
            sb.append(field.name);
            sb.append('=');
            final Object value = record.getFieldValue(i);
            sb.append((value != null) ? value.toString() : "");
        }
    }
    
    private String makeInsertInfo(final MDERecord record, final int retCode) {
        final StringBuffer sb = new StringBuffer(2048);
        sb.append(MDESqlLogger.getLogTime());
        sb.append("\tINSERT\t");
        sb.append(this.getTableName());
        sb.append('\t');
        sb.append(retCode);
        sb.append('\t');
        sb.append(record.getDimValues().toString());
        this.writeFieldValues(record, sb);
        sb.append('\t');
        return sb.toString();
    }
    
    private String makeUpdateInfo(final MDERecord record, final int retCode) {
        final StringBuffer sb = new StringBuffer(2048);
        sb.append(MDESqlLogger.getLogTime());
        sb.append("\tUPDATE\t");
        sb.append(this.getTableName());
        sb.append('\t');
        sb.append(retCode);
        sb.append('\t');
        sb.append(record.getDimValues().toString());
        this.writeFieldValues(record, sb);
        sb.append('\t');
        return sb.toString();
    }
    
    private String makeDeleteInfo(final MDEDimValueSet dimValues, final int retCode) {
        final StringBuffer sb = new StringBuffer(2048);
        sb.append(MDESqlLogger.getLogTime());
        sb.append("\tDELETE\t");
        sb.append(this.getTableName());
        sb.append('\t');
        sb.append(retCode);
        sb.append('\t');
        sb.append(dimValues.toString());
        sb.append('\t');
        return sb.toString();
    }
}
